/*

    2002/05/19 - 1.4 - DeXT
    
    mkvcdfs.c: new option: -x (keep form2 file extension)
    mkvcdfs.c: new file collect routine, to allow subdirs for form2 files
    mkvcdfs.c: default screen output is now stdout (can be redirected)
    mkvcdfs.c: fixed bug where input files were not closed
    mkvcdfs.c: BIN image is now referenced as relative in TOC, CUE files
    mkvcdfs.c: big speedup thanks to new r/w buffering

    2002/05/19 - 1.3 - DeXT
    
    mkvcdfs.c: new options: -a, -d

    2002/05/19 - 1.2 - DeXT
    
    mkvcdfs.c: command line parser is now less tolerant to bad syntax
    mkvcdfs.c: ISO FS is now built before copying actual file data
    mkvcdfs.c: show total estimated CD size
    mkvcdfs.c: new XA subheader which is more compatible with DVD players

    2002/05/06 - 1.2pre2 - DeXT
    
    mkvcdfs.c: added progress counter and speed indicator
    mkvcdfs.c: fixed DAT filesize (was 1 sector higher w/multiple of 2324)
    mkvcdfs.c: fixed CUE file not being deleted on failure
    mkvcdfs.c: XA subheader now marks EOF
    
    2002/05/04 - 1.2pre - Nic & DeXT
    
    mkvcdfs.c: new command line option parser (Nic)
    mkvcdfs.c: new features (set volume name, M2F2 ext, output image) (Nic)
    mkvcdfs.c: changed default volume name and image base name
    mkvcdfs.c: added simple d-char (ISO-9660) converter

    2002/04/18 - beta 1.1 - DeXT
    
    mkvcdfs.c: added CUE output support

    2002/04/17 - beta 1.0 - DeXT

    mkvcdfs.c: changed output file method to fopen() (open() worked in text
      mode, perhaps a MinGW bug?)
    mkvcdfs.c: removed 30+45 empty sector padding in M2F2 track
    mkvcdfs.c: now will accept any file (not just MPEG)
    mkvcdfs.c: changed CD/XA track subheader to Form2/Data
    mkvcdfs.c: fixed DAT file length and track file length:
      * DAT file now has right size (missed 1 sector)
      * CD/XA track now has 2 "ghost" sectors at the end, to
        prevent any potential TAO problems


    -- Original notes --


    mkvcdfs: make a VCD image from several MPEG-1 files suitable
             for burning to CDR with cdrdao

    Usage:

      mkvcdfs mpegfile1 mpegfile2 ....

    mkvcdfs creates 2 files:

    vcd.toc          contains the table of contents of the VCD
    vcd_image.bin    contains the CD-Image itself


    Copyright (C) 2000 Rainer Johanni <Rainer@Johanni.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef _WIN32
#include <windows.h>
#endif
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>

#ifndef WIN32
#include <unistd.h>
#endif

#include "defaults.h"
#include "ecc.h"
#include "buffer.h"

#define EOF_INDICATOR 0xffffffff

static unsigned char data[2324];
static unsigned char outrec[2352];
static int maxrec = 0, prevrec = 0;
FILE *xa_fd;

struct buffer_s read_buffer;
struct buffer_s write_buffer;

int long_file_names;
int ascii;
int form2_ext_to_name;
int num_entries;
int num_dirs;

file_s file[MAX_FILES]; // TODO: allocate memory at runtime

// Extern variables set here
char CD_VOLUME_ID[33];
char BINARY_OUTPUT_FILE[260];
char VCD_TOC_FILE[260];
char VCD_CUE_FILE[260];
char FORM2_EXT[260];
int ISO_FS_BLOCKS = 302;

#ifdef _WIN32
   LARGE_INTEGER Frequency, PerformanceCount, last_counter;   // counter stuff
#endif

static void write_record(int rec)
{
   int n;

   if (rec == prevrec + 1) // consecutive writes only
   {
       n = BufWrite(outrec, 2352, &write_buffer);
   }
   else
   {
       BufWriteFlush(&write_buffer); // flush buffer before seeking
       fseek(xa_fd,rec*2352,SEEK_SET);
       n = fwrite(outrec,2352,1,xa_fd);
   }
   
   prevrec = rec;

   if(n==0)
   {
      fprintf(stderr,"Error writing to binary output file\n");
      perror("write");
      exit(1);
   }
}

static void output_zero(int rec)
{
   /* Output a zero record */

   do_encode_L2(outrec, MODE_0, rec+150);
   write_record(rec);
}

void output_form1(int rec, unsigned char *data)
{
   int i;

   /* Output a CDROM XA Mode 2 Form 1 record,
      we fill gaps with zero records */

   if(rec+1>maxrec)
   {
      for(i=maxrec;i<rec;i++) output_zero(i);
      maxrec = rec+1;
   }

   /* We use the same subheader for all form 1 sectors,
      I dont know if this is completly correct */

   outrec[16] = outrec[20] = 0;
   outrec[17] = outrec[21] = 0;
   outrec[18] = outrec[22] = 8;
   outrec[19] = outrec[23] = 0;

   /* Copy the data to outrec */

   memcpy(outrec+24,data,2048);

   /* Adding of sync, header, ECC, EDC fields */

   do_encode_L2(outrec, MODE_2_FORM_1, rec+150);

   /* Output record */

   write_record(rec);
}

void output_form2(int rec, int h1, int h2, int h3, int h4, unsigned char *data)
{
   int i;

   /* Output a CDROM XA Mode 2 Form 1 record,
      we fill gaps with zero records */

   if(rec+1>maxrec)
   {
      for(i=maxrec;i<rec;i++) output_zero(i);
      maxrec = rec+1;
   }

   /* Subheader */

   outrec[16] = outrec[20] = h1;
   outrec[17] = outrec[21] = h2;
   outrec[18] = outrec[22] = h3;
   outrec[19] = outrec[23] = h4;

   /* Copy the data to outrec */

   memcpy(outrec+24,data,2324);

   /* Adding of sync, header, EDC */

   do_encode_L2(outrec, MODE_2_FORM_2, rec+150);

   /* Output record */

   write_record(rec);
}

int copy_file(file_s file)
{
   FILE *fd;
   int tag, readed;
   unsigned int progress, speed = 0;
   int i, fn, cn, sm, ci, extent;
   int blocksize;

      extent = file.extent;
      
      if (file.type == FORM2)
         blocksize = 2324;
      else // FORM1
         blocksize = 2048;
      
      printf("Copying data from \"%s\"\n", file.name);

      fd = fopen(file.name,"rb");
      if (fd==0)
      {
         fprintf(stderr,"Can not open file %s\n", file.name);
         perror("open");
         return (-1);
      }
      
      read_buffer.file = fd;
      read_buffer.size = READ_BUF_SIZE;
      read_buffer.index = 0;
      read_buffer.ptr = (char *) malloc (READ_BUF_SIZE);
      
      if (read_buffer.ptr == NULL)
      {
         fprintf(stderr,"Not enough free memory for buffers!");
         exit(1);
      }

      /* Output the file itself */

      tag = 0; /* new file starts */

      for(i=0;i<file.size;i++)
      {

#define SHOW_INTERVAL 150
#define SPEED_INTERVAL 1500

         if (!(i%SHOW_INTERVAL))
         {
#ifdef _WIN32
            if (i > 0 && !(i%(SPEED_INTERVAL))) { 
            QueryPerformanceCounter(&PerformanceCount);
            speed = (unsigned int)(SPEED_INTERVAL*blocksize)/(1+(PerformanceCount.QuadPart - last_counter.QuadPart)/(1+(Frequency.QuadPart/1000)));
            last_counter = PerformanceCount;
            }
#endif
            progress = i*100/file.size;
            fprintf(stderr,
                   "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
                   "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
                   "[Progress: %2.2u%%  Speed: %u KB/s]  ",
                   progress, speed);
         }

//         readed = fread(data,1,blocksize,fd);
         BufRead (data, blocksize, &read_buffer, file.length);
         
         if (i == file.size-1) // if last sector
         {
            tag = EOF_INDICATOR;
//            if (readed < blocksize)
//               memset(data+readed, 0, blocksize-readed); // fill remaining with 0s
            if (file.length%blocksize != 0)
               memset(data + (file.length%blocksize), 0, blocksize - (file.length%blocksize)); // fill remaining with 0s
         }

         if(file.type == FORM2 && tag == EOF_INDICATOR && i < 150)  // minimum track size
         {
            fprintf(stderr,"Not enough data\n");
            fclose(fd);
            return (-1);
         }

/*
   XA subheader format: FileNumber, ChannelNumber, SubMode, CodingInformation

   FileNumber = the unique ID for a single XA file
   ChannelNumber = the file is multimedia (1) or not (0)
   SubMode = XA content description: struct SubMode { uint8 EndOfRecord:1;
   CodingInformation = type of data:                  uint8 VideoData:1;
         0x00 = empty/data sector                     uint8 ADCPM:1;
         0x0f = video sector                          uint8 Data:1;
         0x7f = audio sector                          uint8 TriggerOn:1;
         0x80 = MPEG2 (SVCD)                          uint8 Form2:1;
                                                      uint8 RealTime:1;
                                                      uint8 EndOfFile:1; };
*/

         if (file.type == FORM2)
         {
            fn = file.number;
            cn = 1;
            sm = 0x68; // Form2, Data
            ci = 0x80;

            if (tag==EOF_INDICATOR)
            {
               sm |= 0x01; // End of record
               sm |= 0x80; // End of file
            }

            output_form2(extent++,fn,cn,sm,ci,data);
         }   
         else // FORM1
         {
            output_form1(extent++,data);
         }

         if(tag==EOF_INDICATOR)
         {
            fprintf(stderr,
                   "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
                   "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
                   "                                                          "
                   "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
                   "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");

            printf("Done, got %d sectors\n",i+1);
            break;
         }
      }
      
      free(read_buffer.ptr);

      fclose(fd);
      
      return extent;
}

int parse_params(int argc, char **argv)
{
	enum { COMMAND_BAD, COMMAND_SET_VOLUME, COMMAND_SET_OUTPUTNAME, \
		COMMAND_ADD_FORM1_FILES, COMMAND_ADD_MOVIE_FILES, \
		COMMAND_DEFAULT_EXT, COMMAND_ADD_DIR };
	int i;
	int command = COMMAND_ADD_MOVIE_FILES;  // default command
	int form2_file_number = 1;
	
	// Set Defaults First
	strcpy(CD_VOLUME_ID, "MODE2CD");
	strcpy(BINARY_OUTPUT_FILE, "image.bin");
	strcpy(VCD_TOC_FILE, "image.toc");
	strcpy(VCD_CUE_FILE, "image.cue");
	strcpy(FORM2_EXT, "dat");
	long_file_names = 0;
	ascii = 0;
	form2_ext_to_name = 0;
	num_entries = 0;
	num_dirs = 0;
	
	// Loop thru params
	for ( i = 1; i != argc; i ++ )
	{
		if ( argv[i][0] == '-' )	// if argv[i] is a command
		{
			switch ( toupper(argv[i][1]) )
			{
			case '?':
			case 'H':
				return 0;
			case 'V':	// Volume
				command = COMMAND_SET_VOLUME;
				break;
			case 'O':	// output image base name
				command = COMMAND_SET_OUTPUTNAME;
				break;
			case 'F':	// add form1 files
				command = COMMAND_ADD_FORM1_FILES;
				break;
			case 'M':	// add movie files
				command = COMMAND_ADD_MOVIE_FILES;
				break;
			case 'E':   // set M2F2 extension
				command = COMMAND_DEFAULT_EXT;
				break;
			case 'D':   // add dir
				command = COMMAND_ADD_DIR;
				break;
			case 'L':   // long filenames
				long_file_names = 1;
				command = COMMAND_BAD; // next argument has to be a command, not a parameter
				break;
			case 'A':   // ascii filenames
				ascii = 1;
				command = COMMAND_BAD; // next argument has to be a command, not a parameter
				break;
			case 'X':   // copy ext into filename for form2 files
				form2_ext_to_name = 1;
				command = COMMAND_BAD; // next argument has to be a command, not a parameter
				break;
			default:	// if unknown, call bad
				command = COMMAND_BAD;
			}
		}
		else	// if not a command then a param for a command
		{
			switch ( command )
			{
			case COMMAND_BAD:	// unknown command
				return 0;		// FAIL
				break;
			case COMMAND_SET_VOLUME:
				to_dchar(CD_VOLUME_ID, argv[i], 32);
				command = COMMAND_BAD; // next argument has to be a command, not a parameter
				break;
			case COMMAND_SET_OUTPUTNAME:
				strcpy(BINARY_OUTPUT_FILE, argv[i]);
				strcpy(VCD_TOC_FILE, argv[i]);
				strcpy(VCD_CUE_FILE, argv[i]);
				strcat(BINARY_OUTPUT_FILE, ".bin");
				strcat(VCD_TOC_FILE, ".toc");
				strcat(VCD_CUE_FILE, ".cue");
				command = COMMAND_BAD; // next argument has to be a command, not a parameter
				break;
			case COMMAND_ADD_FORM1_FILES:
				if ( num_entries == MAX_FILES )
				{
					fprintf(stderr, "ERROR: Maximum of %d files exceeded!\n",MAX_FILES);
					return 0;	// FAIL
				}
				else
				{
					file[num_entries].name = argv[i];
					file[num_entries].type = FORM1;
					file[num_entries].number = 0;
					num_entries++;
				}
				break;
			case COMMAND_ADD_MOVIE_FILES:
				if ( num_entries == MAX_FILES )
				{
					fprintf(stderr, "ERROR: Maximum of %d files exceeded!\n",MAX_FILES);
					return 0;	// FAIL
				}
				else
				{
					file[num_entries].name = argv[i];
					file[num_entries].type = FORM2;
					file[num_entries].number = form2_file_number++;
					num_entries++;
				}
				break;
			case COMMAND_DEFAULT_EXT:
				strcpy(FORM2_EXT, argv[i]);
				command = COMMAND_BAD; // next argument has to be a command, not a parameter
				break;
			case COMMAND_ADD_DIR:
				if ( num_entries == MAX_FILES || num_dirs == MAX_DIRS)
				{
					fprintf(stderr, "ERROR: Maximum of %d dirs exceeded!\n",MAX_DIRS);
					return 0;	// FAIL
				}
				else
				{
					file[num_entries].name = argv[i];
					file[num_entries].type = DIR;
					num_entries++;
					num_dirs++;
					command = COMMAND_BAD; // next argument has to be a command, not a parameter
				}
				break;
			}
		}
	}

	// Check if any actual input files have been entered
	if ( num_entries - num_dirs == 0)
	{
		fprintf(stderr, "ERROR: No input files have been entered!\n");
		return 0;	// FAIL
	}
	else
		return 1;	// OK
}

void output_usage()
{
	printf("usage:\n");
	printf("mode2cdmaker -m movie1.avi -d dir -f readme.txt -o image -v NAME -e DAT -l -a -x\n\n");
	printf("-h this help\n");
	printf("-v volume name\n");
	printf("-o output image base name (*.bin, *.toc, *.cue)\n");
	printf("-f add Form1 files\n");
	printf("-m add movie (Form2) files (default)\n");
	printf("-d add subdirectory\n");
	printf("-e set default extension for Form2 files\n");
	printf("-l use long file names (ISO-9660 Level 2)\n");
	printf("-a use ASCII character set\n");
	printf("-x keep original Form2 file extension\n");
}

int main (int argc, char **argv)
{
   FILE *fd;
   FILE *fd_toc;
   FILE *fd_cue;
   int i, n, m, s, f;
   int form1_extent, form2_extent, extent, last_form2_file;
   struct stat statbuf;
   char *binfilename;

#ifdef _WIN32
   QueryPerformanceFrequency(&Frequency);     // Init counter
   QueryPerformanceCounter(&last_counter);
#endif

   printf("mode2cdmaker version %s (%s)\n\n", APP_VERSION, APP_DATE);

   /* Parse the parameters & fill in the appropriate variables */

   if ( argc < 2 || !parse_params(argc, argv) )
   {
	   output_usage();
	   exit(1);
   }

   /* Open binary output file */

   binfilename = (char *) strrchr(BINARY_OUTPUT_FILE, DIR_SLASH);
   if (binfilename == NULL)
      binfilename = BINARY_OUTPUT_FILE;
   else
      binfilename = &binfilename[1]; // skip slash

   xa_fd = fopen(BINARY_OUTPUT_FILE,"wb");
   if(xa_fd==0)
   {
      fprintf(stderr,"Can not open %s\n",BINARY_OUTPUT_FILE);
      perror("open");
      exit(1);
   }

   /* Init write buffer */
   
   write_buffer.file = xa_fd;
   write_buffer.size = WRITE_BUF_SIZE;
   write_buffer.index = 0;
   write_buffer.ptr = (char *) malloc (WRITE_BUF_SIZE);

   if (write_buffer.ptr == NULL)
   {
      fprintf(stderr,"Not enough free memory for buffers!");
      exit(1);
   }
 
   /* open VCD_TOC_FILE */

   fd_toc = fopen(VCD_TOC_FILE,"w");
   if(fd_toc==0)
   {
      fprintf(stderr,"Can not open VCD toc file %s\n",VCD_TOC_FILE);
      perror("fopen");
      exit(1);
   }

   /* open VCD_CUE_FILE */

   fd_cue = fopen(VCD_CUE_FILE,"w");
   if(fd_cue==0)
   {
      fprintf(stderr,"Can not open VCD cue file %s\n",VCD_CUE_FILE);
      perror("fopen");
      exit(1);
   }

   /* Collect files */

   printf("Collecting file data...\n");

   form1_extent = START_FILE_EXTENT;
   form2_extent = 0; // relative

   for(n=0;n<num_entries;n++)
   {
      if (file[n].type == DIR) continue; // skip dirs

      fd = fopen(file[n].name,"rb");
      if (fd==0)
      {
         if (stat(file[n].name, &statbuf) == 0)
         {
            file[n].type = NOTFILE; // entry is not a file (skip when building ISO FS)
            continue;
         }
      else
         {   
            fprintf(stderr,"Can not open file %s\n",file[n].name);
            perror("open");
            fprintf(stderr,"Fatal Error --- exiting\n");
            fclose(xa_fd);
            remove(BINARY_OUTPUT_FILE);
            fclose(fd_toc);
            remove(VCD_TOC_FILE);
            fclose(fd_cue);
            remove(VCD_CUE_FILE);
            exit(1);
         }
      }

      fseek(fd, 0L, SEEK_END);
      file[n].length = ftell(fd);
      fseek(fd, 0L, SEEK_SET);
   
      if (file[n].type == FORM2)
      {
         form2_extent += 150; // pregap
         file[n].extent = form2_extent;
         file[n].size = (file[n].length+2323)/2324; // in blocks
         form2_extent += file[n].size+2; // track size

         last_form2_file = n;  // useful for TOC file below
      }
      else if (file[n].type == FORM1)
      {
         file[n].extent = form1_extent;
         file[n].size = LEN2BLOCKS(file[n].length);
         form1_extent += file[n].size;

         if (form1_extent > ISO_FS_BLOCKS)
            ISO_FS_BLOCKS = form1_extent;
      }

      fclose(fd);
   }

   /* Convert FORM2 extents to absolute */

   for(n=0;n<num_entries;n++)
   {
      if (file[n].type == FORM2)
         file[n].extent += ISO_FS_BLOCKS;
   }

   /* Show total CD size */
   
   extent = form2_extent + ISO_FS_BLOCKS; // absolute

   f = extent + 150;
   s = f/75;
   f = f%75;
   m = s/60;
   s = s%60;
   printf("Estimated CD size is %d MB (%2.2d:%2.2d:%2.2d)\n\n", (int)(extent*2048)>>20,m,s,f);

   /* Build the ISO track */

   printf("Building ISO filesystem...\n");

   mk_vcd_iso_fs(file, num_entries, num_dirs, long_file_names, ascii, form2_ext_to_name);
   
   /* Write the toc file */

   fprintf(fd_toc,"CD_ROM_XA\n\n");
   fprintf(fd_toc,"// Track 1: Header with ISO 9660 file system\n");
   fprintf(fd_toc,"TRACK MODE2_RAW\n");
   f = ISO_FS_BLOCKS + 150;
   s = f/75;
   f = f%75;
   m = s/60;
   s = s%60;
   fprintf(fd_toc,"DATAFILE \"%s\" %2.2d:%2.2d:%2.2d\n\n",binfilename,m,s,f);

   /* Write the cue file */
   
   fprintf(fd_cue,"FILE \"%s\" BINARY\n",binfilename);
   fprintf(fd_cue,"  TRACK 01 MODE2/2352\n");
   fprintf(fd_cue,"    INDEX 01 00:00:00\n");

   /* Copy the files */

   printf("\nCopying files...\n");

// copy FORM1 files

   extent = START_FILE_EXTENT;

   for(n=0;n<num_entries;n++)
   {
      if (file[n].type != FORM1) continue; // only FORM1 files

      extent = copy_file(file[n]);
      
      if (extent < 0)
      {
         fprintf(stderr,"Fatal Error --- exiting\n");
         fclose(xa_fd);
         remove(BINARY_OUTPUT_FILE);
         fclose(fd_toc);
         remove(VCD_TOC_FILE);
         fclose(fd_cue);
         remove(VCD_CUE_FILE);
         exit(1);
      }
   }

// fill gap between FORM1 & FORM2 files (if any)

   memset(data,0,2048);
   for(i=extent;i<ISO_FS_BLOCKS;i++)
      output_form1(i,data);

// copy FORM2 files

   extent = ISO_FS_BLOCKS;

   for(n=0;n<num_entries;n++)
   {
      if (file[n].type != FORM2) continue; // only FORM2 files

      /* Pre gap */

      memset(data,0,2324);
      for(i=0;i<150;i++) output_form2(extent++,0,0,0x20,0,data);

      /* Copy files */

      extent = copy_file(file[n]);
      
      if (extent < 0)
      {
         fprintf(stderr,"Fatal Error --- exiting\n");
         fclose(xa_fd);
         remove(BINARY_OUTPUT_FILE);
         fclose(fd_toc);
         remove(VCD_TOC_FILE);
         fclose(fd_cue);
         remove(VCD_CUE_FILE);
         exit(1);
      }
   
      /* postgap (2 empty form 2 blocks at the end) */

      memset(data,0,2324);
      for(i=0;i<2;i++) output_form2(extent++,0,0,0x20,0,data);

      /* Update TOC file */

      fprintf(fd_toc,"// Track %d: data from %s\n",n+2,file[n].name);
      fprintf(fd_toc,"TRACK MODE2_RAW\n");
      f = file[n].size+2; // track size
      if (n != last_form2_file) f += 150;
      s = f/75;
      f = f%75;
      m = s/60;
      s = s%60;
      fprintf(fd_toc,"DATAFILE \"%s\" #%d %2.2d:%2.2d:%2.2d\n\n",
                     binfilename,file[n].extent*2352,m,s,f);

      /* Update CUE file */
      
      f = file[n].extent;
      s = f/75;
      f = f%75;
      m = s/60;
      s = s%60;
      fprintf(fd_cue,"  TRACK %2.2d MODE2/2352\n",n+2);
      fprintf(fd_cue,"    INDEX 01 %2.2d:%2.2d:%2.2d\n",m,s,f);
   }

   BufWriteFlush(&write_buffer); // flush buffer before exiting
   free(write_buffer.ptr);

   fclose(fd_toc);
   fclose(fd_cue);
   fclose(xa_fd);

   exit(0);
}

